/***************************************************** Flash Programmer  ****************************************************

	Red Rapids
	797 North Grove Rd
	Suite 101
	Richardson, TX 75081
	www.redrapids.com | support@redrapids.com

	The following code is used to program the flash on all Red Rapids products.

	Note:	The application can be applied to the maximum number of Red Rapids devices (MAX_NUM_DEVICES) supported, but only
			a single device (DEV0) is demonstrated.

	Rev History:	05-08-2015:	Increased scope to include all four available segments in the flash.
					05-04-2015:	Initial release.

	Copyright 2015, Red Rapids, Inc.

*****************************************************  Flash Programmer  ****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "adapter.h"					// Adpater device driver common to all Red Rapids PCI Express products.
#include "flash.h"						// SigFPGA application.
#include "common_functions.h"
#include "programmer_functions.h"		// Functions used to program the flash device.

int main(int argc, char* argv[])
{
	s_Adapter *p_HW[MAX_NUM_DEVICES];			// Hardware handle to the Red Rapids device.
	unsigned long Status = 0;					// Variable to store status returned by function calls.
	unsigned long Handle[MAX_NUM_DEVICES];		// Array to store hardware handle allocation status for each device (success = 0, failure = 1).
	unsigned long devcnt = 0;					// Variable to index device numbers in loops.
	int index = 0;							// Variable to index array elements in loops.
	char MCSFileName[64] = "No File Specified";	// Character string to store MCS file name.
	s_FlashCode *p_FlashCode = NULL;			// Pointer to the first record in the linked list of flash code.
	s_FlashCode *p_record = NULL;				// Pointer to a flash code record.
	size_t directive = 0;						// Directive requested by a command line argument.
	unsigned long segment = 0;					// Segment of flash requested by a command line argument.
	unsigned long StartAddress = 0;				// Starting address of a requested segment.
	unsigned long EndAddress = 0;				// Ending address of a requested segment.


	printf("***************************** Flash Programmer *****************************\n");

	//////////////////////////////////////  Parse Command Line Arguments And MCS File  ///////////////////////////////////////

	printf("\nChecking command inputs...\n");

	//Proceesing command line arguments.
	for(index = 1; index < argc; index++)
	{
		//Print the command usage to the console if requested with a "help" argument.
		if(!strcmp(argv[index],"-help") || !strcmp(argv[index],"-h") || !strcmp(argv[index],"-?") || !strcmp(argv[index],"?"))
		{
			printf("\nCOMMAND USAGE:\n");
			printf("\nflash [-b -e -u -v] [-p <file> -g <file> -all]\n");
			printf("\nOne of the following four flash segments must be selected:\n");
			printf(" -p\tApply directive to the primary segment.\n");
			printf(" -g\tApply directive to the golden segment.\n");
			printf(" -a1\tApply directive to alternate segment #1.\n");
			printf(" -a2\tApply directive to alternate segment #2.\n");
			printf("\nOne of the following four directives must be selected:\n");
			printf(" -b\t\tBlank check the entire selected segment.\n");
			printf(" -e\t\tErase the entire selected segment.\n");
			printf(" -u <filename>\tUpdate the selected segment <MCS file required>.\n");
			printf(" -v <filename>\tCompare the selected segment <MCS file required>.\n");
			exit(0);
		}
		// Check for more than one directive switch.
		else if(directive && (!strcmp(argv[index],"-b") || !strcmp(argv[index],"-e") || !strcmp(argv[index],"-u") || !strcmp(argv[index],"-v")))
		{
			printf("\tERROR: Only one directive (-b, -e, -u, or -v) allowed.\n");
			exit(1);
		}
		// Set the blank check directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-b"))
			directive = 1;
		// Set the erase directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-e"))
			directive = 2;
		// Set the update directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-u"))
		{
			directive = 3;
			strcpy(MCSFileName, argv[++index]);
		}
		// Set the verify directive flag if the switch is selected.
		else if(!strcmp(argv[index],"-v"))
		{
			directive = 4;
			strcpy(MCSFileName, argv[++index]);
		}
		// Check for more than one segment select switch.
		else if(segment && (!strcmp(argv[index],"-p") || !strcmp(argv[index],"-g") || !strcmp(argv[index],"-a") || !strcmp(argv[index],"-b")))
		{
			printf("\tERROR: Only one segment allowed (-p, -g, -a or -b).\n");
			exit(1);
		}
		// Set the golden segment flag if the switch is selected.
		else if(!strcmp(argv[index],"-g"))
			segment = 0x4ul;
		// Set the primary segment flag if the switch is selected and copy the MCS file name.
		else if(!strcmp(argv[index],"-p"))
			segment = 0x5ul;
		// Set the alternate segment A flag if the switch is selected and copy the MCS file name.
		else if(!strcmp(argv[index],"-a1"))
			segment = 0x6ul;
		// Set the alternate segment B flag if the switch is selected and copy the MCS file name.
		else if(!strcmp(argv[index],"-a2"))
			segment = 0x7ul;
		// Set an error flag if there is an invalid command line argument.
		else 
		{
			printf("\tERROR: Invalid argument entered on the command line (%s).\n", argv[index]);
			exit(1);
		}
	}

	// Check for illegal combinations of command line switches.
	if(!directive)
		{
			printf("\tERROR: At least one directive (-b, -e, -u, or -v) must  be selected.\n");
			exit(1);
		}
	if(!segment)
		{
			printf("\tERROR: At least one segment (-p, -g, -a or -b) must  be selected.\n");
			exit(1);
		}
	if((directive == 3 || directive == 4) && !strcmp(MCSFileName,"No File Specified"))
		{
			printf("\tERROR: The update (-u) or verify (-v) directives require an MCS file.\n");
			exit(1);
		}

	// Parse the MCS file if it was selected.
	if (strcmp(MCSFileName,"No File Specified"))
	{
		// Add the .mcs file extension if it is not inclded in the file name.
		if (strstr(MCSFileName,".mcs") == NULL && strstr(MCSFileName,".MCS") == NULL)
			strcat(MCSFileName,".mcs");
		// Parse the MCS file to extract the flash code.
		p_FlashCode = MCSParser(MCSFileName);
		// Flag an error if a NULL pointer is returned.
		if (p_FlashCode == NULL)
		{
			printf("\tERROR: Parsing the MCS file (%s) failed.\n", MCSFileName);
			exit(1);
		}
	}

	//////////////////////////////////////////////  Initialize Hardware Handle  //////////////////////////////////////////////

	printf("\nInitializing hardware handle...\n");

	// Initialize the hardware handle for each device.
	for (devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
	{
		// Allocate memory for the structure if there were no prior errors.
		if (!Status)
		{
			p_HW[devcnt] = (s_Adapter*) malloc(sizeof(s_Adapter));
			if (p_HW[devcnt] == NULL)
			{
				// This is a fatal error because there is no hardware handle available.
				printf("\tERROR: Unable to allocate memory for the device #%lu hardware handle.\n", devcnt);
				Status++;
				// Mark this handle as a failed allocation.
				Handle[devcnt] = 1;
			}
			else
			{
				// Mark this handle as a successful allocation.
				Handle[devcnt] = 0;
				// Initialize each member of the structure.
				Adapter_Zero(p_HW[devcnt]);
				// Set a unique device number for each hardware handle.
				p_HW[devcnt]->DevNum = devcnt;
			}
		}
		// Mark the handle as a failed memory allocation if there were prior errors.
		else
			Handle[devcnt] = 1;
	}

	//////////////////////////////////////////////////  Configure Hardware  //////////////////////////////////////////////////
	if (!Status)
	{
		printf("\nOpening device #%u on the host...\n", DEV0);

		// Open the Red Rapids device on the bus and only continue if successful.
		if (!BusOpen(p_HW[DEV0]))
		{
			// Print the model number to the screen.
			printf("\tDetected Red Rapids Model %03lX.\n", ModelNumber(p_HW[DEV0]));

			// Print the firmware revision date to the screen.
			printf("\tFirmware revision (MMDDYYYY): %08lX\n", FirmwareRevision(p_HW[DEV0]));

			// Software reset the device.
			Status += SoftwareReset(p_HW[DEV0]);
			
			// Pause 100 useconds for the programming clock to settle.
			MicrosecondTimer(100);

			// Configure the programmer if the clock is stable.
			if (GetClockStatus(p_HW[DEV0]))
			{
				printf("\tProgramming clock stable.\n");

				// Enable the flash chip.
				Status += ProgEnable(p_HW[DEV0]);

				// Set the selected segment.
				Status += ProgSegment(p_HW[DEV0], segment);

				// Initialize the flash chip.
				Status += ProgInit(p_HW[DEV0]);
			}
			// Report an error if the programming clock is not stable.
			else
			{
				printf("\tERROR: Programming clock not stable.\n");
				Status++;
			}

	////////////////////////////////////////////  Execute Programmer Directives  //////////////////////////////////////////////

			if (!Status)
			{

				printf("\nWARNING: The following operations may take several minutes to complete.");
				printf("\nWARNING: Any interruption may require an external programmer to recover!\n");

				// Set the starting and ending address.
				if (directive == 3 || directive == 4)
				{
					// Find the last record in the flash code.
					p_record = p_FlashCode;
					while (p_record->next != NULL)
						p_record = p_record->next;

					// Set the address range equal to the flash code.
					StartAddress = p_FlashCode->address;
					EndAddress = p_record->address;
				}
				else
				{
					// Set the address range equal to an entire segment.
					StartAddress = 0x0ul;
					EndAddress = (BLOCK_SIZE*BLOCK_COUNT/SEGMENT_COUNT) - BLOCK_SIZE;
				}
					

				// Perform blank check operation.
				if (directive == 1)
				{
					printf("\nPerforming blank check operation...\n");
					if (BlankCheck(p_HW[DEV0], StartAddress, EndAddress))
						printf("\tSegment is not blank.\n");
					else
						printf("\tSegment is blank.\n");
				}

				// Perform erase operation.
				if (directive == 2 || directive == 3)
				{
					printf("\nPerforming unlock operation...\n");
					// Unlock the selected address range.
					Status += Unlock(p_HW[DEV0], StartAddress, EndAddress);
					if (!Status)
					{
						// Unlock completed without errors.
						printf("\tSegment is unlocked.\n");

						printf("\nPerforming erase operation...\n");
						// Erase the selected address range.
						Status += Erase(p_HW[DEV0], StartAddress, EndAddress);
						if (!Status)
							// Erase completed without errors.
							printf("\tSegment is erased.\n");
						else
						{
							// Erase operation reported errors.
							printf("\tERROR: Segment erase failed.\n");
							Status++;
						}
					}
					else
					{
							// Unlock operation reported errors.
							printf("\tERROR: Segment unlock failed.\n");
							Status++;
					}
				}

				// Perform program operation.
				if (directive == 3 && !Status)
				{
					printf("\nPerforming program operation...\n");
					Status += Program(p_HW[DEV0], p_FlashCode);
					if (!Status)
						// Programming completed without errors.
						printf("\tSegment is programmed.\n");
					else
						// Programming operation reported errors.
						printf("\tERROR: Segment program failed.\n");
				}
	
				// Perform verify operation.
				if (directive == 4)
				{
					printf("\nPerforming verify operation...\n");
					if (Verify(p_HW[DEV0], p_FlashCode))
						printf("\tSegment does not match MCS file.\n");
					else
						printf("\tSegment matches MCS file.\n");
				}

				printf("\n...all operations completed.\n");

				// Disable the flash chip.
				Status += ProgDisable(p_HW[DEV0]);

				// Close the Red Rapids device on the bus.
				Status += BusClose(p_HW[DEV0]);
			}
			else
				printf("\n...terminating execution due to hardware configuration errors.\n");
		}
		else
			printf("\n...terminating execution because device could not be opened.\n");
	}
	else
		printf("\n...terminating execution due to hardware handle intialization errors.\n");

	/////////////////////////////////////////////////  Free Allocated Memory /////////////////////////////////////////////////

	// Free the memory allocated to the hardware handles.
	for (devcnt = 0; devcnt < MAX_NUM_DEVICES; devcnt++)
		if (!Handle[devcnt]) free(p_HW[devcnt]);

	// Free the memory allocated to the flash program code.
	while (p_FlashCode != NULL)
	{
		// Save the location of the next record in the linked list.
		p_record = p_FlashCode->next;
		// Free the current record in the linked list.
		free(p_FlashCode);
		// Move to the next record in the linked list.
		p_FlashCode = p_record;
	}

return 0;
}
